import logging
from typing import Any, Dict, List, Optional

import requests

from livy.models import Version, Session, SessionKind, Statement, StatementKind


LOGGER = logging.getLogger(__name__)


VALID_LEGACY_SESSION_KINDS = {
    SessionKind.SPARK, SessionKind.PYSPARK, SessionKind.PYSPARK3,
    SessionKind.SPARKR
}
VALID_SESSION_KINDS = {
    SessionKind.SPARK, SessionKind.PYSPARK, SessionKind.SPARKR,
    SessionKind.SQL, SessionKind.SHARED
}


class JsonClient:

    def __init__(self, url: str) -> None:
        self.url = url
        self.session = requests.Session()

    def close(self) -> None:
        self.session.close()

    def get(self, endpoint: str='') -> dict:
        return self._request('GET', endpoint)

    def post(self, endpoint: str, data: dict=None) -> dict:
        return self._request('POST', endpoint, data)

    def delete(self, endpoint: str='') -> dict:
        return self._request('DELETE', endpoint)

    def _request(
        self, method: str, endpoint: str, data: dict=None
    ) -> dict:
        url = self.url.rstrip('/') + endpoint
        response = self.session.request(method, url, json=data)
        response.raise_for_status()
        return response.json()


class LivyClient:

    def __init__(self, url: str) -> None:
        self._client = JsonClient(url)
        self._server_version_cache: Optional[Version] = None

    def close(self) -> None:
        self._client.close()

    def server_version(self) -> Version:
        if self._server_version_cache is None:
            data = self._client.get('/version')
            self._server_version_cache = Version(data['version'])
        return self._server_version_cache

    def legacy_server(self) -> bool:
        version = self.server_version()
        return version < Version('0.5.0-incubating')

    def list_sessions(self) -> List[Session]:
        data = self._client.get('/sessions')
        return [Session.from_json(item) for item in data['sessions']]

    def create_session(
            self, kind: SessionKind, spark_conf: Dict[str, Any]=None
    ) -> Session:
        if self.legacy_server():
            valid_kinds = VALID_LEGACY_SESSION_KINDS
        else:
            valid_kinds = VALID_SESSION_KINDS

        if kind not in valid_kinds:
            raise ValueError(
                f'{kind} is not a valid session kind for a Livy server of '
                f'this version (should be one of {valid_kinds})'
            )

        body = {'kind': kind.value}
        if spark_conf is not None:
            body['conf'] = spark_conf

        data = self._client.post('/sessions', data=body)
        return Session.from_json(data)

    def get_session(self, session_id: int) -> Optional[Session]:
        try:
            data = self._client.get(f'/sessions/{session_id}')
        except requests.HTTPError as e:
            if e.response.status_code == 404:
                return None
            else:
                raise
        return Session.from_json(data)

    def delete_session(self, session_id: int) -> None:
        self._client.delete(f'/sessions/{session_id}')

    def list_statements(self, session_id: int) -> List[Statement]:
        response = self._client.get(f'/sessions/{session_id}/statements')
        return [
            Statement.from_json(session_id, data)
            for data in response['statements']
        ]

    def create_statement(
        self, session_id: int, code: str, kind: StatementKind=None
    ) -> Statement:

        data = {'code': code}

        if kind is not None:
            if self.legacy_server():
                LOGGER.warning('statement kind ignored on Livy<0.5.0')
            data['kind'] = kind.value

        response = self._client.post(
            f'/sessions/{session_id}/statements',
            data=data
        )
        return Statement.from_json(session_id, response)

    def get_statement(
        self, session_id: int, statement_id: int
    ) -> Statement:
        response = self._client.get(
            f'/sessions/{session_id}/statements/{statement_id}'
        )
        return Statement.from_json(session_id, response)



import re
from enum import Enum
from functools import total_ordering
from typing import NamedTuple, Optional, List


@total_ordering
class Version:

    def __init__(self, version: str) -> None:
        match = re.match(r'(\d+)\.(\d+)\.(\d+)(\S*)$', version)
        if match is None:
            raise ValueError(f'invalid version string {version!r}')
        self.major, self.minor, self.dot, self.extension = match.groups()

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return f'{name}({self.major}.{self.minor}.{self.dot}{self.extension})'

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Version) and
            self.major == other.major and
            self.minor == other.minor and
            self.dot == other.dot
        )

    def __lt__(self, other: 'Version') -> bool:
        if self.major < other.major:
            return True
        elif self.major == other.major:
            if self.minor < other.minor:
                return True
            elif self.minor == other.minor:
                return self.dot < other.dot
            else:
                return False
        else:
            return False


class SparkRuntimeError(Exception):

    def __init__(
        self, ename: Optional[str], evalue: Optional[str],
        traceback: Optional[List[str]]
    ) -> None:
        self.ename = ename
        self.evalue = evalue
        self.traceback = traceback

    def __repr__(self) -> str:
        name = self.__class__.__name__
        components = []
        if self.ename is not None:
            components.append(f'ename={self.ename!r}')
        if self.evalue is not None:
            components.append(f'evalue={self.evalue!r}')
        return f'{name}({", ".join(components)})'


class OutputStatus(Enum):
    OK = 'ok'
    ERROR = 'error'


_Output = NamedTuple(
    '_Output',
    [
        ('status', OutputStatus),
        ('text', Optional[str]),
        ('json', Optional[dict]),
        ('ename', Optional[str]),
        ('evalue', Optional[str]),
        ('traceback', Optional[List[str]])
    ]
)


class Output(_Output):

    @classmethod
    def from_json(cls, data: dict) -> 'Output':
        return cls(
            OutputStatus(data['status']),
            data.get('data', {}).get('text/plain'),
            data.get('data', {}).get('application/json'),
            data.get('ename'),
            data.get('evalue'),
            data.get('traceback')
        )

    def raise_for_status(self) -> None:
        if self.status == OutputStatus.ERROR:
            raise SparkRuntimeError(self.ename, self.evalue, self.traceback)


class StatementKind(Enum):
    SPARK = 'spark'
    PYSPARK = 'pyspark'
    SPARKR = 'sparkr'
    SQL = 'sql'


class StatementState(Enum):
    WAITING = 'waiting'
    RUNNING = 'running'
    AVAILABLE = 'available'
    ERROR = 'error'
    CANCELLING = 'cancelling'
    CANCELLED = 'cancelled'


_Statement = NamedTuple(
    '_Statement',
    [
        ('session_id', int),
        ('statement_id', int),
        ('state', StatementState),
        ('output', Optional[Output])
    ]
)


class Statement(_Statement):

    @classmethod
    def from_json(cls, session_id: int, data: dict) -> 'Statement':
        if data['output'] is None:
            output = None
        else:
            output = Output.from_json(data['output'])
        return cls(
            session_id,
            data['id'],
            StatementState(data['state']),
            output
        )


class SessionKind(Enum):
    SPARK = 'spark'
    PYSPARK = 'pyspark'
    PYSPARK3 = 'pyspark3'
    SPARKR = 'sparkr'
    SQL = 'sql'
    SHARED = 'shared'


class SessionState(Enum):
    NOT_STARTED = 'not_started'
    STARTING = 'starting'
    IDLE = 'idle'
    BUSY = 'busy'
    SHUTTING_DOWN = 'shutting_down'
    ERROR = 'error'
    DEAD = 'dead'
    SUCCESS = 'success'


_Session = NamedTuple(
    '_Session',
    [('session_id', int), ('kind', SessionKind), ('state', SessionState)]
)


class Session(_Session):

    @classmethod
    def from_json(cls, data: dict) -> 'Session':
        return cls(
            data['id'],
            SessionKind(data['kind']),
            SessionState(data['state'])
        )
